# This file is part of CloudControl.
#
# CloudControl is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CloudControl is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with CloudControl.  If not, see <http://www.gnu.org/licenses/>.


import datetime
from Queue import Queue, Empty

#TODO: implement iterable interface
#TODO: implement priority responses

class AsyncWatcher(object):

    """ Asynchronous call watcher -- Handle asynchronous calls and responses.

    Usage example::

      >>> watcher = AsyncWatcher()
      >>> watcher.register(conn1, 'ping')
      >>> watcher.register(conn2, 'ping')
      >>> responses = watcher.wait() # Wait for all responses
    """

    def __init__(self):
        self._response_queue = Queue()
        self._expected_responses = {}

    def _get_in_queue(self, *args, **kwargs):
        """ Get an item in response queue and return it with the time waited to
            its comming. Accept all arguments accepted by :meth:`Queue.get`
        """
        start = datetime.datetime.now()
        item = self._response_queue.get(*args, **kwargs)
        dt = datetime.datetime.now() - start
        dt = dt.seconds + dt.days * 86400 + dt.microseconds * 0.000001
        return (dt, item)

#
# Public API:
#

    @property
    def remains(self):
        """ Remaining expected responses.
        """
        return len(self._expected_responses)

    def register(self, rpc, method, *args, **kwargs):
        """ Call specified method on specified rpc with specified arguments
            and register the call on this object.

        :param rpc: the rpc on which made the call
        :param method: the rpc method to call
        :param \*args, \*\*kwargs: the arguments to pass to the rpc method
        """
        msg_id = rpc.async_call(self._response_queue, method, *args, **kwargs)
        self._expected_responses[msg_id] = kwargs.pop('_data', None)
        # Data are stored directly on the AsyncWatcher instead of RpcProtocol.

    def wait(self, timeout=None, max_wait=None, raise_timeout=False):
        """ Wait responses for calls registered on this :class:`AsyncWatcher`
            instance and return them.

        :param timeout: timeout value or None to disable timeout (default: None)
        :param max_wait: maximum waited responses
        :return: received messages in a list

        .. note::

           You can repeat call to this method on a single
           :class:`AsyncWatcher`. For example, if you want to process the
           first arrived message, then wait others for a minute, you can do::

             >>> msgs = watcher.wait(max_wait=1)
             >>> process_speedy(msgs[0])
             >>> for msg in watcher.wait(timeout=60):
             >>>     process(msg)
        """
        return list(self.iter(timeout, max_wait, raise_timeout))

    def iter(self, timeout=None, max_wait=None, raise_timeout=False):
        """ Work like :meth:`AsyncWatcher.wait` but return an iterable, instead
            of a list object.

        .. note::

           Responses are yielded by the iterable when they are received, so you
           can start the processing of response before receiving all.
        """

        while self.remains:
            try:
                dt, response = self._get_in_queue(timeout=timeout)
            except Empty:
                break
            if timeout is not None:
                timeout -= dt
            response['data'] = self._expected_responses[response['id']]
            del self._expected_responses[response['id']]
            yield response
            # Check for max_wait:
            if max_wait is not None:
                max_wait -= 1
                if not max_wait:
                    break
        if timeout is not None and raise_timeout:
            for expected, data in self._expected_responses.iteritems():
                yield {'error': {'exception': 'TimeoutError',
                                 'message': 'Response not received on time'},
                       'data': data,
                       'id': expected}
            self._expected_responses = {}
